/**
 * Funktioner för att underkänna kravavsnitt från EN 301 549.
 *
 * För varje kontroll i mnaualen skapar skriptet ett verktygsfält med knappar
 * för att underkänna avsnittet (för aktuellt granskningsobjekt eller flera) och
 * möjlighet att koppla taggar till globalt underkända avsnitt.
 *
 * Skriptet presenterar en rapport med alla underkända krav i slutet av sidan,
 * tillsammans med knappar för att återställa lokala underkännanden eller
 * samtliga underkännanden och taggar. Rapporten indikerar vilka krav användaren
 * underkänt globalt och vilket som eventuellt undantagits lokalt.
 *
 * I början av sidan visar skriptet en lista med alla definierade taggar, där
 * användaren kan avmarkera taggar som inte är relevanta för det aktuella
 * granksningsobjektet. Tillhörande globala underkännanden blir då automatiskt
 * lokalt godkända.
 *
 * Detta låter användaren fokusera på granskningen och på att notera
 * anmärkningar i granskningsprotokollet. När webbsidan, dokumentet eller
 * skärmbilden granskats färdigt kan användaren använda lista med
 * underkända krav för att uppdatera motsvarande kravflik i protokollet
 * och sedan rensa för att återställa inför nästa granskningsobjekt.
 *
 * Information om underkända avsnitt sparas i persistenslagret (cookies), med
 * användarens medgivande.
 *
 * Klasser
 * -------
 * Clause     Kravavsnitt med information om underkännanden.
 * Report     Rapport över underkända avsnitt.
 * ReportItem Ett element i rapporten över underkända avsnitt.
 * Storage    Lagring i persistenslagret.
 * Tags       Lista med kryssrutor för använda taggar.
 * Toolbar    Verktygsfält fÖr ett kravavsnitt.
 */
((doc) => {
    "use strict";


    /**
     * Ett kravavsnitt i EN 301 549.
     *
     * Klassen innehåller en vektor med avsnitt och tillhandahåller en statisk
     * metod för att hämta en instans för ett specifikt avsnitt. Man skapar
     * alltså inte instanser med `new`.
     *
     * Varje instans bevakar de verktygsfält som är kopplade till avsnittet, och
     * håller reda på underkännandestatus för avsnittet.
     *
     * @see Toolbar
     * @property {boolean} failed Sant om avsnittet är underkänt.
     * @property {string}  label  Avsnittsnummer, till exempel "9.1.2.3".
     * @property {boolean} many   Sant om avsnittet är globalt underkänt.
     * @property {boolean} one    Sant om avsnittet är lokalt underkänt.
     */
    class Clause {

        static _clauses      = [];
        static _anyCallbacks = [];

        /**
         * Skapa och returnera en instans.
         *
         * Instansen lagras i en statisk vektor och bara en instans existerar
         * per kravavsnitt.
         *
         * Klienter ska hämta instanser med den här metoden, inte skapa egna
         * instanser med `new Clause()`.
         *
         * @param {string} label Avsnittsnummer.
         */
        static forClause (label) {
            // Omvandla avsnittsnumret till ett numeriskt index.
            const index = label.split(".")
                .map((c) => parseInt(c, 10))
                .concat([0, 0])
                .slice(0, 4)
                .reduce((a, c) => 100 * a + c, 0);

            if (undefined === Clause._clauses[index]) {
                Clause._clauses[index] = new Clause(label);
            }
            return Clause._clauses[index];
        }

        /**
         * Återställ alla underkännanden.
         */
        static resetHard () {
            Clause._clauses.forEach((clause) => clause._resetHard());
        }

        /**
         * Återställ alla lokala underkännanden.
         */
        static resetSoft () {
            Clause._clauses.forEach((clause) => clause._resetSoft());
        }

        /**
         * Returnera ett löfte om meddelande vid förändring.
         *
         * Löftet infrias så snart något kravavsnitt ändrar status.
         *
         * @return {Promise}
         */
        static whenAnyChange () {
            return new Promise((resolve, reject) => {
                Clause._anyCallbacks.push(() => resolve());
            });
        }

        /**
         * Konstruera en avsnittsinstans.
         *
         * @private
         * @param {string} label Avsnittsnummer.
         */
        constructor (label) {
            this._callbacks = [];
            this._label     = label;
            this._many      = [];
            this._one       = [];
        }

        // Åtkomstmetoder.
        static get clauses () {return Clause._clauses}

        get failed () {return this.one || this.many}
        get label  () {return this._label          }
        get many   () {return this._many.length > 0}
        get one    () {return this._one .length > 0}

        /**
         * Bevaka ett verktygsfält.
         *
         * Begär att bli meddelad om verktygsfältet status förändras, det vill
         * säga om användaren underkänner eller godkänner kravavsnittet.
         *
         * @param {Toolbar} toolbar Verktygsfält.
         */
        watchToolbar (toolbar) {
            toolbar.whenStateChanges()
            .then((toolbar) => {
                const many = this._many.indexOf(toolbar);
                const one  = this._one. indexOf(toolbar);

                if (-1 === one && toolbar.one) {
                    // Nytt lokalt underkännande.
                    this._one.push(toolbar);
                } else if (-1 !== one && !toolbar.one) {
                    // Nytt lokalt godkännande.
                    this._one.splice(one, 1);
                }

                if (-1 === many && toolbar.many) {
                    // Nytt globalt underkännande.
                    this._many.push(toolbar);
                } else if (-1 !== many && !toolbar.many) {
                    // Nytt globalt godkännande.
                    this._many.splice(many, 1);
                }

                return toolbar;

            }).then((toolbar) => {
                this._resolvePromises();
                return toolbar;

            }).then((toolbar) => this.watchToolbar(toolbar));
        }

        /**
         * Returnera ett löfte om meddelande vid förändring.
         *
         * Löftet infrias så snart ett verktygsfält underkänner eller godkänner
         * kravavsnittet.
         *
         * @return {Promise}
         */
        whenChanged () {
            return new Promise((resolve, reject) => {
                this._callbacks.push(() => resolve(this));
            });
        }

        /**
         * Återställ lokala och globala underkännanden.
         *
         * @private
         */
        _resetHard () {
            this._many.forEach((toolbar) => toolbar.resetHard());
            this._one .forEach((toolbar) => toolbar.resetHard());
            this._many = [];
            this._one  = [];
            this._resolvePromises();
        }

        /**
         * Återställ lokala underkännanden.
         *
         * @private
         */
        _resetSoft () {
            this._many.forEach((toolbar) => toolbar.resetSoft());
            this._one .forEach((toolbar) => toolbar.resetSoft());
            this._one = Array.from(this._many);
            this._resolvePromises();
        }

        /**
         * Infria givna löften om meddelande vid förändring.
         *
         * @see whenChanged()
         */
        _resolvePromises () {
            this._callbacks     .forEach((callback) => callback());
            Clause._anyCallbacks.forEach((callback) => callback());
            this._callbacks      = [];
            Clause._anyCallbacks = [];
        }

    }


    /**
     * Rapport över underkända kravklausuler.
     *
     * Repporten representeras av en UL-lista med ett LI-element per
     * avsnittsnummer. Varje LI-element representeras av en instans av klassen
     * `FailureReportItem`.
     *
     * @see ReportItem
     * @property {Clause[]} clauses Kravavsnitt.
     */
    class Report {

        /**
         * Konstruera en listinstans.
         *
         * Ersätter platshållaren med en SECTION innehållande en H2-rubrik, en
         * UL-lista och två rensaknappar. Lägger även in en ankarlänk till
         * rubriken i sidnavigeringen.
         *
         * @param {string}        heading     Rubriktext.
         * @param {HTMLElement}   placeholder Platshållare.
         * @param {HTMLLIElement} navSibling  Syskonelement i navigeringen.
         */
        constructor (heading, placeholder, navSibling) {
            const section   = this._replacePlaceholderWithSection(placeholder);
            this._addHeading(section, heading);
            this._ul        = this._addList(section);
            this._buttons   = [];
            this._callbacks = [];
            this._clauses   = [];

            const p = section.appendChild(doc.createElement("p"));
            this._buttons.push(this._addButton(
                p,
                "Återställ",
                "Återställ lokala underkännanden",
                () => Clause.resetSoft()
            ));
            this._buttons.push(this._addButton(
                p,
                "Nollställ helt",
                "Återställ lokala och globala underkännanden",
                () => Clause.resetHard(),
                true
            ));

            this._addLinkToPageNav(heading, navSibling);
        }

        // Åtkomstmetoder.
        set clauses (clauses) {
            this._clauses = clauses;
            this._clauses.forEach((clause) => new ReportItem(this._ul, clause));

            const watchClauses = () => Clause.whenAnyChange()
            .then(() => this._enableResetButtons())
            .then(() => watchClauses());

            watchClauses();
        }

        /**
         * Returnera ett löfte om meddelande vid återställning.
         *
         * Löftet infrias när användaren klickar på någon av återställnings-
         * knapparna.
         *
         * @return {Promise}
         */
        whenReset () {
            return new Promise((resolve, reject) => {
                this._callbacks.push((type) => resolve(type));
            });
        }

        /**
         * Lägg till en återställningsknapp.
         *
         * @private
         * @param {HTMLElement} parent  Föräldraelement.
         * @param {string}      label   Knapptext.
         * @param {string}      title   Tooltip-text.
         * @param {function}    onClick Callbackfunktion vid klick.
         * @param {boolean}     [hard]  Hård återställning?
         * @return {HTMLButtonElement}
         */
        _addButton (parent, label, title, onClick, hard = false) {
            const button    = parent.appendChild(doc.createElement("button"));
            button.type     = "button";
            button.disabled = true;
            button.title    = title;
            if (hard) {
                button.className = "secondary";
            }
            button.appendChild(doc.createTextNode(label));
            button.addEventListener("click", (event) => {
                event.stopPropagation();
                event.preventDefault();
                const type = hard ? "hard" : "soft";
                onClick();
                this._callbacks.forEach((callback) => callback(type));
            });
            return button;
        }

        /**
         * Lägg till en H2-rubrik.
         *
         * @private
         * @param {HTMLElement} parent  Föräldraelement.
         * @param {string}      heading Rubriktext.
         */
        _addHeading (parent, heading) {
            const h2    = parent.appendChild(doc.createElement("h2"));
            h2.tabIndex = 0;
            h2.id       = "fail-section";
            h2.appendChild(doc.createTextNode(heading));
        }

        /**
         * Infoga ankarlänk till rapporten i sidnavigeringen.
         *
         * @private
         * @param {string}        text    Länktext.
         * @param {HTMLLIElement} sibling Syskonelement.
         */
        _addLinkToPageNav (text, sibling) {
            const link = sibling.parentNode
            .insertBefore(doc.createElement("li"), sibling)
            .appendChild(doc.createElement("a"));
            link.href  = "#fail-section";
            link.appendChild(doc.createTextNode(text));
        }

        /**
         * Lägg till en UL-lista.
         *
         * @private
         * @param {HTMLElement} parent Föräldraelement.
         * @return {HTMLUListElement}
         */
        _addList (parent) {
            return parent.appendChild(doc.createElement("UL"));
        }

        /**
         * Aktivera eller inaktivera återställningsknapparna.
         *
         * Om inget element i rapporten är synligt ska knapparna vara
         * inaktiverade.
         *
         * @private
         */
        _enableResetButtons () {
            const visible = this._clauses.some((clause) => clause.failed);
            this._buttons.forEach((button) => button.disabled = !visible);
        }

        /**
         * Ersätt platshållaren i dokumentet med ett SECTION-element.
         *
         * @private
         * @param {HTMLElement} placeholder Platshållare.
         * @return {HTMLElement}
         */
        _replacePlaceholderWithSection (placeholder) {
            const section     = doc.createElement("section");
            section.className = "failed-requirements";
            placeholder.parentNode.replaceChild(section, placeholder);
            return section;
        }

    }


    /**
     * Ett element i rapporten över underkända avsnitt.
     *
     * Motsvarar ett LI-element i en oordnad lista. Med `class`- och `hidden`-
     * attribut håller klassen reda på underkännandestatus och styr
     * presentationen. Kombinationerna är som följer:
     *
     * Lokalt underkänd Globalt underkänd class    hidden
     * ---------------- ----------------- -----    ------
     * ja               nej               local    false
     * nej              ja                override false
     * ja               ja                         false
     * nej              nej                        true
     *
     * @see Report
     */
    class ReportItem {

        /**
         * Konstruera en elementinstans.
         *
         * @param {HTMLUListElement} list   UL-lista att infoga elementet i.
         * @param {Clause}           clause Kravavsnitt.
         */
        constructor (list, clause) {
            this._li        = list.appendChild(doc.createElement("li"));
            this._li.hidden = true;
            this._li.appendChild(doc.createTextNode(clause.label));

            const watchClause = (clause) => clause.whenChanged()
            .then((clause) => this._update(clause))
            .then((clause) => watchClause(clause));

            watchClause(clause);
        }

        /**
         * Uppdatera LI-elementet när underkännandestatus förändrats.
         *
         * @private
         * @param {Clause} clause Kravavsnitt.
         * @return {Clause}
         */
        _update (clause) {
            const many = clause.many;
            const one  = clause.one;

            this._li.hidden = !(one || many);
            this._li.className = !many ? "local" : (!one ? "override" : "");

            return clause;
        }
    }


    /**
     * Persistent lagring av underkännanden.
     *
     * Persistenslagret består av en cookie (`manual.X.fail` per typ av
     * granskningsobjekt (webbsida, dokument, mobil applikation).
     *
     * Klassen sätter inga cookies utan användarens uttryckliga medgivande (som
     * persisteras med cookien `manual.cookies`.
     *
     * @property {boolean} enabled Sant om användaren samtyckt till kakor.
     */
    class Storage {

        /**
         * Konstruera en lagringsinstans.
         *
         * Hämta lagrade underkännanden från persistenslagret.
         *
         * @param {string} key  Lagringsnyckel: web/doc/app.
         * @param {number} days Lagringstid (dagar).
         */
        constructor (key, days) {
            this._days     = days;
            this._enabled  = /(^|\s*)manual\.cookies=allow(;|$)/
                             .test(doc.cookie);
            this._fails    = {};
            this._key      = key;
            this._toolbars = [];

            if (this._enabled) {
                this._read();
            }
        }

        // Åtkomstmetoder.
        get enabled () {return this._enabled}

        /**
         * Koppla ett verktygsfält till lagringsinstansen.
         *
         * Om någon av knapparna sparats i lagret blir den nedtryckt.
         *
         * @param {Toolbar} toolbar Verktygsfält.
         */
        addToolbar (toolbar) {
            this._toolbars.push(toolbar);

            // Om verktygsfältet sparats i persistenslagret.
            if (toolbar.id in this._fails) {
                const fail = this._fails[toolbar.id];
                toolbar.populate(fail.many, fail.tags);

            } else if (this._legacy) {
                // Kompatibilitet ned cookies skapade av äldre skriptversioner.
                const index = this._toolbars.length - 1;
                if (`${index}` in this._legacy.global) {
                    toolbar.populate(
                        true,
                        this._legacy.global[`${index}`].split(" ")
                    );
                } else if (-1 !== this._legacy.local.indexOf(index)) {
                    toolbar.populate(false, []);
                }
            }

            // Bevaka ändringar i verktygsfältet.
            const watchToolbarState = (toolbar) => {
                toolbar.whenStateChanges()
                .then((toolbar) => {
                    this._write();
                    return toolbar;
                }).then((toolbar) => watchToolbarState(toolbar));
            };
            watchToolbarState(toolbar);

            const watchToolbarTags = (toolbar) => {
                toolbar.whenTagsChange()
                .then(([toolbar, event, tags]) => {
                    this._write();
                    return toolbar;
                }).then((toolbar) => watchToolbarTags(toolbar));
            };
            watchToolbarTags(toolbar);

        }

        /**
         * Aktivera persistent lagring efter användaren gett sitt samtycke.
         */
        enable () {
            const date = new Date();
            date.setTime(date.getTime() + 365 * 24 * 60 * 60 * 1000);
            const expires = date.toUTCString();

            this._enabled = true;
            doc.cookie   = `manual.cookies=allow;expires=${expires};path=/`;
        }

        /**
         * Radera alla underkännanden i persistenslagret.
         *
         * Detta ska bara ske vid "hård" återställning.
         *
         * @param {string} type Typ av återställning (hard/soft).
         * @return {string}
         */
        reset (type) {
            if ("hard" === type && this._enabled) {
                const date = (new Date(0)).toUTCString();
                const suffix = `expires=${date};path=/;secure`;
                doc.cookie = `manual.${this._key}.fail=;${suffix}`;
                doc.cookie = `manual.${this._key}.local=;${suffix}`;
                doc.cookie = `manual.${this._key}.global=;${suffix}`;
            } else {
                this._write();
            }
        }

        /**
         * Läs och tolka information från persistenslagret.
         *
         * Resultatet lagras i propertyn `_fails` med respektive verktygsfälts
         * ID som nycklar.
         */
        _read () {
            const start   = "(?:^|;\\s*)";
            const key     = `manual\\.${this._key}\\.fail`;
            const data    = "([^;]+)";
            const end     = "(?:;|$)";
            const matches = RegExp(`${start}${key}=${data}${end}`)
                            .exec(doc.cookie);

            if (matches) {
                matches[1].split(",").map((data) => data.split(":"))
                .forEach(([id, level, tags]) => {
                    this._fails[id] = {
                        many: ("G" === level),
                        tags: (tags || "").split("|")
                              .filter((tag) => "" !== tag)
                              .map   ((tag) => decodeURIComponent(tag))
                    };
                });

            } else {
                this._readLegacy();
            }
        }

        /**
         * Läs och tolka kakor skapade med en äldre version av skriptet.
         */
        _readLegacy () {
            if (!/(^|;\s*)manual\.[^.]+\.(global|local)=/.test(doc.cookie)) {
                return;
            }

            const conv  = (index) => index + (index >= 40 ? 1 : 0);
            const parse = (level) => {
                const start   = "(?:^|;\\s*)";
                const key     = `manual\\.${this._key}\\.${level}`;
                const data    = "([^;]+)";
                const end     = "(?:;|$)";
                const matches = RegExp(`${start}${key}=${data}${end}`)
                                .exec(doc.cookie);
                return matches ? matches[1].split(",") : [];
            };

            this._legacy        = {};
            this._legacy.global = {};
            parse("global").map((data) => data.split(":", 2))
            .forEach(([index, tags]) => {
                this._legacy.global[conv(parseInt(index, 10))] =
                    decodeURIComponent(tags);
            });

            this._legacy.local = parse("local")
                                 .map((index) => conv(parseInt(index, 10)));
        }

        /**
         * Skriv verktygsfältsinformation till persistenslagret.
         *
         * Sätt en cookie med namnet "manual.X.fail", där X är aktuell typ av
         * granskningsobjekt (app/doc/wwb). Värdet består av en kommaseparerad
         * lista av underkännanden. Varje underkännande innehåller
         * kolonseparerade komponenter: identitet, nivå (L/G för "lokalt" eller
         * "globalt") och, för globala underkännanden, en lista med taggar
         * åtskilda av vertikala streck.
         *
         * @example
         * manual.web.fail=foo9123:L,bar9234:G:abc|def
         *
         * Exemplet visar två verktygsfält. Ett, med id "foo9123" har ett lokalt
         * underkännande. Det andra, med id "bar9234" har ett globalt (och
         * därmed lokalt) underkännande och två kopplade taggar: "abc" och
         * "def".
         */
        _write () {
            if (!this._enabled) {
                return;
            }

            // Komprimera underkännanden och taggar till en sträng.
            const level = (tb) => tb.many ? `G:${tags(tb)}` : "L";
            const tags  = (tb) => tb.tags.map((t) => encodeURIComponent(t));
            const value = this._toolbars
                          .filter((tb) => tb.failed)
                          .map   ((tb) => `${tb.id}:${level(tb)}`)
                          .join  (",");

            // Bäst-före-datum.
            const date = new Date();
            date.setTime(
                "" === value
                ? 0
                : date.getTime() + this._days * 24 * 60 * 60 * 1000
            );

            // Spara cookie.
            const key = `manual.${this._key}.fail`;
            doc.cookie = `${key}=${value};expires=${date.toUTCString()}`
                       + ";path=/;secure";

            // Radera eventuella cookies från äldre skriptversioner.
            if (this._legacy) {
                date.setTime(0);
                const suffix = `expires=${date.toUTCString()};path=/;secure`;
                doc.cookie   = `manual.${this._key}.local=;${suffix}`;
                doc.cookie   = `manual.${this._key}.global=;${suffix}`;
            }
        }

    }


    /**
     * Företeckning över använda taggar.
     *
     * Hanterar ett DATALIST-element kopplat till inamtningsfältet för taggar i
     * varje verktygsfält.
     *
     * Visar också en lista med definierade taggar med en kryssruta för varje
     * tagg. Om användaren avmarkerar en tagg uppdateras motsvarande
     * underkännanden som kan markeras som lokalt godkänt om alla kopplade
     * taggar avmarkerats.
     */
    class Tags {

        /**
         * Konstruera en instans.
         *
         * @param {HTMLElement}         placeholder Platshållare.
         * @param {HTMLDataListElement} datalist    DATALIST-element.
         * @param {string}              label       Etikett.
         * @throws {Error}
         */
        constructor (placeholder, datalist, label) {
            if (!(placeholder instanceof HTMLElement)) {
                console.log(placeholder);
                throw new Error("Placeholder must be an HTML element.");
            }
            if (!(datalist instanceof HTMLDataListElement)) {
                console.log(datalist);
                throw new Error("Datalist must be a <datalist> element.");
            }

            this._callbacks    = [];
            this._datalist     = datalist;
            this._tags         = {};
            this._ul           = doc.createElement("ul");
            this._ul.className = "tag-reset";
            placeholder.parentNode.replaceChild(this._ul, placeholder);
            this._ul.setAttribute("aria-label", label);
        }

        /**
         * Återställ listan över taggar.
         *
         * @param {string} type Typ av återställning (hard/soft).
         * @return {string}
         */
        reset (type) {
            switch (type) {
                case "hard":
                    this._datalist.innerHTML = "";
                    this._ul.innerHTML       = "";
                    this._tags               = {};
                    break;
                case "soft":
                    this._ul.querySelectorAll("input")
                    .forEach((cb) => cb.checked = true);
                    break;
                default:
                    break;
            }
            return type;
        }

        /**
         * Lägg till övervakning av ett verktygsfält.
         *
         * @param {Toolbar} toolbar Verktygsfält att övervaka.
         */
        watchToolbar (toolbar) {
            toolbar.whenTagsChange()
            .then(([toolbar, event, tags]) => {
                switch (event) {
                    case "add":
                        this._addTags(toolbar, tags);
                        break;
                    case "remove":
                        this._removeTags(toolbar, tags);
                        break;
                    default:
                        break;
                }
                return toolbar;
            }).then((toolbar) => this.watchToolbar(toolbar));
        }

        /**
         * Lägg till taggar.
         *
         * Lägg till nya taggar, om de inte redan finns.
         * Koppla verktygsfältet till taggarna.
         *
         * @param {Toolbar}  toolbar Verktygsfält.
         * @param {string[]} tags    Taggar.
         */
        _addTags (toolbar, tags) {
            tags.forEach((tag) => {
                if (!(tag in this._tags)) {
                    this._tags[tag] = [];
                }

                if (-1 === this._tags[tag].indexOf(toolbar)) {
                    this._tags[tag].push(toolbar);
                }
            });

            this._updateDataList();
            this._updateTagsList();
        }

        /**
         * Ta bort ett verktygsfält från taggar.
         *
         * Om inga andra verktygsfält är kopplade till en tagg, ta bort taggen.
         *
         * @param {Toolbar}  toolbar Verktygsfält.
         * @param {string[]} tags    Taggar.
         */
        _removeTags (toolbar, tags) {
            tags.forEach((tag) => {
                if (tag in this._tags) {
                    const index = this._tags[tag].indexOf(toolbar);
                    if (-1 !== index) {
                        this._tags[tag].splice(index, 1);
                    }

                    if (0 === this._tags[tag].length) {
                        delete this._tags[tag];
                    }
                }
            });

            this._updateDataList();
            this._updateTagsList();
        }

        /**
         * Uppdatera DATALIST-elementet efter uppdatering av taggar.
         *
         * @private
         */
        _updateDataList () {
            this._datalist.innerHTML = "";
            Object.keys(this._tags).sort().forEach((tag) => {
                this._datalist.appendChild(new Option(tag));
            });
        }

        /**
         * Uppdatera listan med avmarkerbara taggar efter uppdatering.
         *
         * @private
         */
        _updateTagsList() {
            this._ul.innerHTML = "";
            Object.entries(this._tags)
            .sort(([[key1, val1], [key2, val2]]) => key1.localeCompare(key2))
            .forEach(([tag, toolbars], index) => {
                const li    = this._ul.appendChild(doc.createElement("li"));

                const cb   = li.appendChild(doc.createElement("input"));
                cb.type    = "checkbox";
                cb.id      = `tag-${index}`;
                cb.checked = toolbars.length > 0;
                cb.value   = tag;

                const label   = li.appendChild(doc.createElement("label"));
                label.htmlFor = cb.id;
                label.appendChild(doc.createTextNode(tag));

                cb.addEventListener("input", (event) => {
                    const target = event.target;
                    const method = target.checked ? "enable" : "disable";
                    this._tags[target.value]
                    .forEach((toolbar) => toolbar[method](target.value));
                });
            });
        }

    }


    /**
     * Verktygsfält för att underkänna eller godkänna kravavsnitt.
     *
     * Verktygsfältet består av
     * - en knapp för att underkänna eller godkänna ett kravavsnitt för ett
     *   specifikt granskningsobjekt ("lokalt")
     * - en knapp för att underkänna eller godkänna ett kravavsnitt för flera
     *   eller alla granskningsobjekt ("globalt")
     * - ett inmatningsfält för taggar kopplade till globala underkännanden
     * - en knapp för att lägga till en tagg
     * - en knapp för varje tillagd tagg.
     *
     * @property {boolean}  failed Sant om avsnittet är underkänt.
     * @property {string}   id     Unik identitet för persistens.
     * @property {boolean}  one    Sant om avsnittet är lokalt underkänt.
     * @property {boolean}  many   Sant om avsnittet är globalt underkänt.
     * @property {string[]} tags   Kopplade taggar.
     */
    class Toolbar {

        /**
         * Konstruera en verktygsfältsinstans.
         *
         * Föräldraelementet måste ha attributet `data-id` som ska innehålla en
         * unik identitet för verktygsfältet. Identiteten används vid persistent
         * lagring, för att särskilja verktygsfält kopplade till samma
         * kravavsnitt. Det måste också ha attributet `data-req` som ska
         * innehålla kravavsnittsnumret, till exempel "9.1.2.3".
         *
         * @param {HTMLElement}         parent  Föräldraelement.
         * @param {HTMLDataListElement} tagList Datalista för taggar.
         * @throws {Error}
         */
        constructor (parent, tagList) {
            if (!(parent instanceof HTMLElement)) {
                console.log(parent);
                throw new Error("Parent must be an HTML element.");
            }
            if (undefined === parent.dataset.id) {
                console.log(parent);
                throw new Error("Parent element must have data-id attribute.");
            }
            if (undefined === parent.dataset.req) {
                console.log(parent);
                throw new Error("Parent element must have data-req attribute.");
            }
            if (!(tagList instanceof HTMLDataListElement)) {
                console.log(tagList);
                throw new Error("Tag list must be a <datalist> element.");
            }
            if ("" === tagList.id) {
                console.log(tagList);
                throw new Error("Tag list must have an id attribute.");
            }

            this._focused        = null;
            this._id             = parent.dataset.id;
            this._manualOverride = false;
            this._many           = null;
            this._one            = null;
            this._stateCallbacks = [];
            this._tagButton      = null;
            this._tagCallbacks   = [];
            this._tagInput       = null;
            this._tags           = {};
            this._toolbar        = this._createToolbarContainer(parent);
            this._createFailContainer(this._toolbar, tagList);
            this._tagContainer   = this._createTagContainer(this._toolbar);
        }

        // Åtkomstmetoder.
        get failed () {return this.many || this.one}
        get id     () {return this._id}
        get many   () {return "true" === this._many.getAttribute("aria-pressed")}
        get one    () {return "true" === this._one .getAttribute("aria-pressed")}
        get tags   () {return Object.keys(this._tags)}

        /**
         * Inaktivera en tagg.
         *
         * Inaktivera motsvarande taggknapp. Om ingen aktiv tagg återstår,
         * sätt den lokala knappen till "inte nedtryckt", om inte användaren
         * manuellt ändrat knappens tillstånd.
         *
         * @param {string} tag Tagg att inaktivera.
         */
        disable (tag) {
            if (tag in this._tags) {
                this._tags[tag].disabled = true;
                if (!this._manualOverride) {
                    if (Object.entries(this._tags)
                        .every(([key, button]) => button.disabled)
                    ) {
                        this._one.setAttribute("aria-pressed", "false");
                        this._callStateCallbacks();
                    }
                }
            }
        }

        /**
         * Aktivera en tagg.
         *
         * Aktivera motsvarande taggknapp. Sätt den lokala knappen till
         * "nedtryckt", om inte användaren manuellt ändrat knappens tillstånd.
         *
         * @param {string} tag Tagg att aktivera.
         */
        enable (tag) {
            if (tag in this._tags) {
                this._tags[tag].disabled = false;
                if (!this._manualOverride && this.many) {
                    this._one.setAttribute("aria-pressed", "true");
                    this._callStateCallbacks();
                }
            }
        }

        /**
         * Populera instansen med externa data.
         *
         * Externa data kan till exempel ha lästs från persistenslagret.
         *
         * @param {boolean}  many Sann om globalt underkänt.
         * @param {string[]} tags Kopplade taggar.
         */
        populate (many, tags) {
            this._many.setAttribute("aria-pressed", String(many));
            this._one .setAttribute("aria-pressed", "true");
            this._tagInput .disabled  = false;
            this._tagButton.disabled  = false;
            this._tagContainer.hidden = false;

            tags.forEach((tag) => this._addTag(tag));

            this._callStateCallbacks();
            if (tags.length > 0) {
                this._callTagCallbacks("add", tags);
            }
        }

        /**
         * Gör en "hård" återställning av verktygsfältet.
         *
         * Sätt båda underkännandeknapparnas tillstånd till "inte nedtryckt",
         * inaktivera inmatningsfält och knapp för taggar samt ta bort alla
         * taggknappar. Anropa _inte_ lyssarfunktioner.
         *
         * @see resetSoft()
         */
        resetHard () {
            this._many.setAttribute("aria-pressed", "false");
            this._one .setAttribute("aria-pressed", "false");
            this._manualOverride         = false;
            this._tagInput .value        = "";
            this._tagInput .disabled     = true;
            this._tagButton.disabled     = true;
            this._tagContainer.innerHTML = "";
            this._tags                   = {};
        }

        /**
         * Gör en "mjuk" återställning av verktygsfältet.
         *
         * Sätt den lokala knappens tillstånd till detsamma som den globala
         * knappens. Anropa _inte_ lyssnarfunktioner.
         *
         * @see resetHard()
         */
        resetSoft () {
            this._manualOverride = false;
            this._one.setAttribute(
                "aria-pressed",
                this._many.getAttribute("aria-pressed")
            );
        }

        /**
         * Informera när någon underkännandeknapps tillstånd förändras.
         *
         * Returnera ett löfte som infrias när användaren ändrar tillståndet på
         * den lokala eller globala knappen. Löftet infrias inte vid
         * återställning av verktygsfältet.
         *
         * @return {Promise}
         */
        whenStateChanges () {
            return new Promise((resolve, reject) => {
                this._stateCallbacks.push(() => resolve(this));
            });
        }

        /**
         * Informera vid tillägg och borttag av taggar.
         *
         * Returnera ett löfte som infrias när användare lägger till eller tar
         * bort taggar. Löftet infrias inte vid återställning av
         * verktygsfältet.
         *
         * @return {Promise}
         */
        whenTagsChange () {
            return new Promise((resolve, reject) => {
                this._tagCallbacks
                .push((event, tags) => resolve([this, event, tags]));
            });
        }

        /**
         * Lägg till en taggknapp i taggbehållaren.
         *
         * @private
         * @param {string} tag Tagg att lägga till.
         */
        _addTag (tag) {
            if (tag in this._tags) {
                return;
            }
            const button    = doc.createElement("button");
            button.type     = "button";
            button.tabIndex = -1;
            button.appendChild(doc.createElement("span"))
            .appendChild(doc.createTextNode(tag));
            button.addEventListener("click", (event) => {
                event.stopPropagation();
                event.preventDefault();
                this._removeTag(button);
            });
            this._tagContainer.appendChild(button);
            this._tags[tag] = button;
            this._callTagCallbacks("add", tag);
        }

        /**
         * Anropa lyssnarfunktioner för knapptillstånd.
         *
         * @see whenStateChanges()
         * @private
         */
        _callStateCallbacks () {
            this._stateCallbacks.forEach((callback) => callback());
            this._stateCallbacks = [];
        }

        /**
         * Anropa lyssnarfunktioner för tagghändelser.
         *
         * @see whenTagsChange()
         * @private
         * @param {string} event Händelse (add/remove).
         * @param {string} tag   Tagg.
         */
        _callTagCallbacks (event, ...tag) {
            this._tagCallbacks.forEach((callback) => callback(event, tag));
            this._tagCallbacks = [];
        }

        /**
         * Skapa en knapp för att underkänna eller godkänna ett kravavsnitt.
         *
         * @private
         * @param {HTMLFormElement} form
         * @param {string}          label
         * @return {HTMLButtonElement}
         */
        _createButton (form, label) {
            // Skapa knappen.
            const button    = form.appendChild(doc.createElement("button"));
            button.type     = "button";
            button.tabIndex = -1;
            button.setAttribute("aria-pressed", "false");

            // Lägg till tomt SPAN för visuell ikon.
            button.appendChild(doc.createElement("span"))
            .setAttribute("aria-hidden", "true");

            // Lägg till dolt span för uppläsande hjälpmedel.
            const span     = button.appendChild(doc.createElement("span"));
            span.className = "s";
            span.appendChild(doc.createTextNode(label));

            return button;
        }

        /**
         * Skapa behållaren för underkännanden och godkännanden.
         *
         * Behållaren är ett formulär med två knappar (lokalt/globalt) och ett
         * inmatningsfält för taggar.
         *
         * @private
         * @param {HTMLDivElement}      toolbar Verktygsfältsbehållaren.
         * @param {HTMLDataListElement} tagList Datalista för taggar.
         */
        _createFailContainer (toolbar, tagList) {
            const form  = toolbar.appendChild(doc.createElement("form"));
            this._one   = this._createButton(form, "lokalt underkänt");
            this._many  = this._createButton(form, "globalt underkänt");
            this._createTagInput(form, tagList);

            this._one .addEventListener("click", () => this._toggleOne ());
            this._many.addEventListener("click", () => this._toggleMany());
        }

        /**
         * Skapa behållaren för taggknappar.
         *
         * @private
         * @param {HTMLDivElement} toolbar Verktygsfältsbehållaren.
         * @return {HTMLDivElement}
         */
        _createTagContainer (toolbar) {
            return toolbar.appendChild(doc.createElement("div"));
        }

        /**
         * Skapa inmatningsfält för taggar, med tillhörande knapp.
         *
         * @private
         * @param {HTMLFormElement}     form    Formulär.
         * @param {HTMLDataListElement} tagList Datalista för taggar.
         */
        _createTagInput (form, tagList) {
            // Inmatningsfält.
            this._tagInput = form.appendChild(doc.createElement("input"));
            this._tagInput.type     = "text";
            this._tagInput.tabIndex = -1;
            this._tagInput.disabled = true;
            this._tagInput.setAttribute("list",       tagList.id);
            this._tagInput.setAttribute("aria-label", "tagg");
            this._tagInput.addEventListener("keydown", (event) => {
                if ("Tab" === event.key && !event.altKey && !event.ctrlKey
                    && !event.metaKey
                ) {
                    event.stopPropagation();
                    event.preventDefault();
                    if (event.shiftKey) {
                        event.target.previousElementSibling.focus();
                    } else {
                        event.target.nextElementSibling.focus();
                    }
                }
            });

            // Knapp.
            this._tagButton = form.appendChild(doc.createElement("button"));
            this._tagButton.tabIndex = -1;
            this._tagButton.disabled = true;
            this._tagButton.appendChild(doc.createTextNode("+"));
            form.addEventListener("submit", (event) => {
                event.stopPropagation();
                event.preventDefault();
                const tag = this._tagInput.value.trim();
                if ("" !== tag) {
                    this._addTag(tag);
                    this._tagInput.value = "";
                }
            });
        }

        /**
         * Skapa behållaren för verktygsfältet.
         *
         * Behållaren är en DIV med rollen "toolbar".
         *
         * @private
         * @param {HTMLElement} parent Föräldraelement.
         * @return {HTMLDivElement}
         */
        _createToolbarContainer (parent) {
            const div     = doc.createElement("div");
            div.setAttribute("role",      "toolbar");
            div.setAttribute("aria-label", parent.dataset.req);

            // Flytta befintligt innehåll från föräldern till DIV:en. Lägg sedan
            // till DIV:en som barn till föräldern.
            div.innerHTML    = parent.innerHTML;
            parent.innerHTML = "";
            parent.appendChild(div);

            div.addEventListener("focus",   (ev) => this._focus(ev), true);
            div.addEventListener("keydown", (ev) => this._keyboard(ev));

            return div;
        }

        /**
         * Håll reda på vilket element som har flytande fokus.
         *
         * @private
         * @param {FocusEvent} event Händelseinformation.
         */
        _focus (event) {
            if (null !== this._focused) {
                this._focused.tabIndex = -1;
            }
            this._focused = event.target;
            this._focused.tabIndex = 0;
        }

        /**
         * Tangentbordsinteraktion i verktygsfältet.
         *
         * Vänster- och högerpil navigerar till föregående respektive nästa
         * element, förutom i inmatningsfältet fÖr taggar. Uppåt- och nedåtpil
         * har samma funktioner och fungerar även i inmatningsfältet.
         *
         * Home och End navigerar till första respektive sista elementet.
         *
         * @private
         * @param {KeyboardEvent} event Händelseinformation.
         */
        _keyboard (event) {
            if (null === this._focused || "INPUT" === this._focused.tagName
                || event.altKey || event.ctrlKey || event.metaKey
                || event.shiftKey
            ) {
                return;
            }

            const focusable = Array.prototype.slice.call(
                this._toolbar.querySelectorAll(
                    "a, button:not(:disabled), input:not(:disabled)"
                )
            );
            let   index  = focusable.indexOf(this._focused);
            const length = focusable.length;

            switch (event.key) {
                case "ArrowLeft":
                case "ArrowUp":
                    index = (--index + length) % length;
                    break;

                case "ArrowRight":
                case "ArrowDown":
                    index = ++index % length;
                    break;

                case "Home":
                    index = 0;
                    break;

                case "End":
                    index = length - 1;
                    break;

                default:
                    return;
            }

            event.stopPropagation();
            event.preventDefault();
            focusable[index].focus();
            this._focused = focusable[index];
        }

        /**
         * Ta bort en taggknapp ur behållaren.
         *
         * @private
         * @param {HTMLButtonElement} button Taggknapp att ta bort.
         */
        _removeTag (button) {
            const tag = button.innerText;
            delete this._tags[tag];
            this._tagContainer.removeChild(button);
            this._callTagCallbacks("remove", tag);
        }

        /**
         * Växla tillstånd på knappen för globalt underkännande.
         *
         * Sätt samma tillstånd på knappen för lokalt underkännande.
         * Om det nya tillståndet är "nedtryckt", gör inmatningsfältet aktivera
         * tagg och tillhörande knapp; i annat fall, inaktivera dem.
         *
         * @private
         */
        _toggleMany () {
            const pressed = "false" === this._many.getAttribute("aria-pressed");
            this._many.setAttribute("aria-pressed", String(pressed));
            this._one .setAttribute("aria-pressed", String(pressed));
            this._tagInput .disabled  = !pressed;
            this._tagButton.disabled  = !pressed;
            this._tagContainer.hidden = !pressed;
            this._callStateCallbacks();
        }

        /**
         * Växla tillstånd på knappen för lokalt underkännande.
         *
         * @private
         */
        _toggleOne () {
            const pressed = "false" === this._one.getAttribute("aria-pressed");
            this._one.setAttribute("aria-pressed", String(pressed));
            this._manualOverride = false;
            this._callStateCallbacks();
        }

    }


    /*
     * Initiering.
     */

    // Kontrollera kompatibilitet.
    if (   "undefined" === typeof doc.querySelectorAll
        || "undefined" === typeof NodeList.prototype.forEach
    ) {
        return;
    }

    // Persistent lagring av underkännanden och taggar.
    const storage = new Storage(doc.body.dataset.persistence, 30);

    // Tagglista.
    const tags = new Tags(
        doc.getElementById("tag-reset"),
        doc.getElementById("tags-list"),
        "Taggar"
    );

    // Skapa verktygsfält för varje kontroll.
    const datalist = doc.getElementById("tags-list");
    doc.querySelectorAll("dd[data-req]").forEach((dd) => {
        const clause  = Clause.forClause(dd.dataset.req);
        const toolbar = new Toolbar(dd, datalist);
        clause  .watchToolbar(toolbar);
        tags    .watchToolbar(toolbar);
        storage.addToolbar   (toolbar);
    });

    // Lista med underkända kravavsnitt.
    const report = new Report(
        "Underkända avsnitt",
        doc.getElementById("failed-requirements"),
        doc.querySelector("main > nav ul > li:last-child")
    );
    report.clauses = Clause.clauses;

    const watchReset = (report) => {
        report.whenReset()
        .then((type) => tags.reset(type))
        .then((type) => storage.reset(type))
        .then    (() => watchReset(report));
    };
    watchReset(report);

    // Infoga cookieinformation om användaren inte lämnat sitt medgivande.
    if (!storage.enabled && !doc.getElementById("cookies")) {
        const p     = doc.querySelector("body > header")
                      .appendChild(doc.createElement("p"));
        p.id        = "cookies";
        p.innerHTML = `För att kunna spara färger och underkända avsnitt
behöver du tillåta kakor <i lang="en">(cookies)</i>. `;

        const allow = p.appendChild(doc.createElement("button"));
        allow.type  = "button";
        allow.appendChild(doc.createTextNode("Tillåt"));
        allow.addEventListener("click", (event) => {
            event.preventDefault();
            event.stopPropagation();
            storage.enable();
            p.parentNode.removeChild(p);
        });
    }

    // Infoga åtgärdspunkter under Efterarbete.
    let ol     = doc.querySelector("#postproc + details ol");
    const last = ol.childElementCount > 1 ? ol.lastElementChild : null;
    ol.insertBefore(doc.createElement("li"), last)
    .appendChild(doc.createElement("p"))
    .innerHTML = `Återställ lokalt underkända avsnitt
(knappen <samp>Återställ</samp>).`;

    ol = doc.querySelector("#postproc + details + details ol");
    if (ol) {
        ol.appendChild(doc.createElement("li"))
        .appendChild(doc.createElement("p"))
        .innerHTML = `Nollställ lokalt och globalt underkända avsnitt
(knappen <samp>Nollställ helt</samp>).`;
    }
})(document);
